unit STI_TMNU;

interface

Uses
   CRT,                                     { standard DOS unit             }
   DOS,                                     { stndard CRT unit              }
   STI_STRN,                                { string handling unit          }
   STI_SCRF,                                { screen functions unit         }
   STI_KEYS;                                { keyboard reading unit         }

Const
  STI_NOERROR          =  0;                { no errors occurred            }
  STI_NOMENUFILE       = -1;                { unable to access menu specs.  }
  STI_FILEREADERROR    = -2;                { unable to read definition     }
  STI_MODECLASH        = -3;                { wrong file for menu mode      }



function STI_GetMenuChoice(PassOn : boolean; Var ExitFlag : boolean; Var Inch : KeyRec) : word;
function STI_MenusInit(MenuFile            : string;
                       HelpRoutine,
                       ValidationRoutine,
                       ExitRoutine,
                       LinkedMenuData      : pointer) : shortint;

implementation

{------------------- Internal Data Structures ------------------------------}

Const
  MAXSTRLEN            = 60;                { maximum size of prompt/title  }
  MAXDEFSIZE           = 32000;             { max. size of memory buffer    }
  MAXSEL               = 512;               { max. number of selections     }
  MAXSTACK             = 512;               { max. menu stack size          }

  NULLCMD              = 0;                 { no command                    }
  UPCMD                = 1;                 { move up command               }
  DOWNCMD              = 2;                 { move down command             }
  LEFTCMD              = 3;                 { move left command             }
  RIGHTCMD             = 4;                 { move right command            }
  SELECTCMD            = 5;                 { select command (return)       }
  ESCAPECMD            = 6;                 { escape command (escape key)   }
  HELPCMD              = 7;                 { help or F1 pressed            }

Type
  BigString = array[0..65534] of char;      { useful for debugging          }
  BigBuffer = array[0..65534] of byte;      { a big kind of buffer          }
  Head = record                             { menu file header              }
           Name      : string[MAXSTRLEN];   { the name of the file          }
           Version   : word;                { the version number            }
           Mode      : byte;                { graphics/text mode            }
           Save      : boolean;             { save the original screen flag }
           BackColor : byte;                { background color              }
           InputS    : byte;                { input style                   }
           NumMenus  : word;                { the number of menus           }
         end;
  Sel  = record                             { one selection                 }
           Prompt    : string[MAXSTRLEN];   { the prompt                    }
           Jump      : boolean;             { should we jump or return      }
           Value     : word;                { return value or next menu     }
         end;
  Posy = record                             { record to hold buffer pos.    }
           MenuID    : word;                { menu identifier               }
           Position  : longint;             { position in the buffer        }
         end;
  Ent  = record                             { one menu definition record    }
           MenuID    : word;                { the menu identifier           }
           MenuTitle : string[MAXSTRLEN];   { the menu title                }
           MenuType  : byte;                { the type of menu (BAR/BOX)    }
           BorderType: byte;                { the type of border            }
           BodyCol   : byte;                { the color of the menu body    }
           BorderCol : byte;                { the border color              }
           TextCol   : byte;                { the text color                }
           HighLight : byte;                { the highlight color           }
           PromptCol : byte;                { the prompt color              }
           TitleCol  : byte;                { the title color               }
           Size      : array[1..4] of byte; { screen coordinates            }
           SelectNum : word;                { the number of selections      }
         end;
  SelA = array[1..MAXSEL] of Sel;           { array of selections           }
  SelF = array[1..MAXSEL] of boolean;       { array of verification flags   }

  OneMenuPtr = ^OneMenu;                    { pointer to one menu           }
  OneMenu    = record                       { one menu's record             }
                 Win   : WindowSave;        { pointer to a saved window     }
                 X,Y   : byte;              { current cursor positions      }
                 CurSel: word;              { current selections            }
                 SS,SE : word;              { first and last sels displayed }
                 Menu  : Ent;               { the menu definition           }
                 Sels  : ^SelA;             { the selections                }
                 SelF  : ^SelF;             { the verifiaction flags        }
                 OnScr,                     { already displayed flag        }
                 ASel  : boolean;           { auto selection flag           }
               end;
  HelpProcType   = procedure(MenuNum,SelNum : word);
  VerFuncType    = function(MenuNum,SelNum : word) : boolean;
  ExitFuncType   = function(MenuNum,SelNum : word) : boolean;

Var
  MenuDefinition : ^BigBuffer;              { pointer to the definition     }
  MenuDefSize    : longint;                 { size of the definition        }
  MenuHeader     : head;                    { pointer to the header         }
  MenuPosy       : posy;                    { one single position buffer    }
  MenuSel        : sel;                     { one selection                 }
  MenuEnt        : ent;                     { one menu entry                }

  InFile         : file;                    { the input file                }
  BufferPos      : word;                    { position in buffer            }
  InMemory       : boolean;                 { is the buffer in memory       }
  Menus          : array[1..MAXSTACK] of OneMenuPtr; { the menu stack       }
  CurrentMenu    : word;                    { number of the current menu    }

  HelpProc       : HelpProcType;            { the help procedure            }
  VerFunc        : VerFuncType;             { the verification procedure    }
  ExitFunc       : ExitFuncType;            { the exit procedure            }
  OldWin         : WindowSave;              { the old screen                }

{---------------------------------------------------------------------------}

procedure FatalError(Message : string);     { fatal error handler           }

begin
  TextColor(White);                         { reset the colors              }
  TextReverse(NoReverse);                   { reset the attribute           }
  writeln;
  writeln(Message);                         { write the message             }
  writeln;
  halt;                                     { die                           }
end;

{---------------------------------------------------------------------------}

procedure ResetColors;                      { restore colors to the default }

begin
  TextColor(White);                         { which is white and            }
  TextReverse(NoReverse);                   { no reverse                    }
end;

{---------------------------------------------------------------------------}

procedure SeekBuffer(Pos : word);           { move to a pos in buffer/file  }

begin
  if InMemory then                          { is it in memory ?             }
    BufferPos := Pos                        { yes, set the buffer pos       }
  else
    seek(InFile,Pos);                       { no, set the file pointer      }
end;

{---------------------------------------------------------------------------}

procedure BlockReadBuffer(Var Buffer; NoBytes : word); { read data          }

begin
  if InMemory then                          { is it in memory               }
    begin
      move(MenuDefinition^[BufferPos],      { yes, copy over the data       }
           Buffer,
           NoBytes);
      inc(BufferPos,NoBytes);               { increment the buffer pointer  }
      if BufferPos > MenuDefSize then       { check for subscript error     }
        begin
          FatalError('Error : Read past buffer end.');
        end;
    end
  else
    begin
      blockread(InFile,Buffer,NoBytes);     { no, read in the data          }
      if IOResult <> 0 then
        begin
          FatalError('Error : File read error.');
        end;
    end;
end;

{---------------------------------------------------------------------------}

function GetBType(InType : byte) : string;  { return border type            }

begin
  case InType of
    0  : GetBType := NOBORDER;              { no border                     }
    1  : GetBType := SPACES;                { use spaces                    }
    2  : GetBType := SINGLELINE;            { straight single line          }
    3  : GetBType := ROUNDCORNERSINGLE;     { single line with round corners}
    4  : GetBType := BIGBLOCK;              { graphics big blocks           }
    5  : GetBType := THICKTOPTHINSIDES;     { as it says                    }
    6  : GetBType := THICKDIAGONALCORNER;   { big block with diagonals      }
    else GetBType := SINGLELINE;            { default type                  }
  end;{case}
end;

{---------------------------------------------------------------------------}

function GetCType(Col : byte) : byte;       { return the color type         }

begin
  Case Col of
    0   : GetCType := BLACK;
    1   : GetCType := BLUE;
    2   : GetCType := GREEN;
    3   : GetCType := CYAN;
    4   : GetCType := RED;
    5   : GetCType := MAGENTA;
    6   : GetCType := YELLOW;
    7,8 : GetCType := WHITE;
    9   : GetCType := BLACK+REVERSE;
    10  : GetCType := BLUE+REVERSE;
    11  : GetCType := GREEN+REVERSE;
    12  : GetCType := CYAN+REVERSE;
    13  : GetCType := RED+REVERSE;
    14  : GetCType := MAGENTA+REVERSE;
    15  : GetCType := YELLOW+REVERSE;
    16  : GetCType := WHITE+REVERSE;
    else  GetCType := WHITE;
  end;
end;

{---------------------------------------------------------------------------}

procedure ReadMenu(Var MenuPtr : OneMenuPtr); { read a menu from buffer     }

Var
  Loop : word;

begin
  if MenuPtr <> NIL then                    { can't read over initialised   }
    FatalError('Error : Reading a menu over an initialised pointer');
  new(MenuPtr);
  BlockReadBuffer(MenuPtr^.Menu,            { read in the menu              }
                  sizeof(Ent));
  getmem(MenuPtr^.SelS,                     { get selection memory          }
         MenuPtr^.Menu.SelectNum * sizeof(Sel));
  getmem(MenuPtr^.SelF,                     { get verification memory       }
         Menuptr^.Menu.SelectNum);
  for Loop := 1 to MenuPtr^.Menu.SelectNum do { loop over the selections    }
    begin
      BlockReadBuffer(MenuPtr^.SelS^[Loop], { raed in the selection         }
                      sizeof(Sel));
      MenuPtr^.SelF^[Loop] := TRUE;         { set flag to true (default)    }
    end;
  MenuPtr^.X        := 1;                   { changed later                 }
  MenuPtr^.Y        := 1;                   { changed later                 }
  MenuPtr^.CurSel   := 1;                   {      "                        }
  MenuPtr^.SS       := 1;                   {      "                        }
  MenuPtr^.SE       := 1;                   {      "                        }
  MenuPtr^.ASel     := FALSE;               {      "                        }
  MenuPtr^.OnScr    := FALSE;               {      "                        }
end;

{---------------------------------------------------------------------------}

procedure SeekMenu(MenuNum : word);         { move to a menus position      }

begin
  SeekBuffer(sizeof(Head)+((MenuNum-1)*sizeof(Posy))); { seek the pos rec.  }
  BlockReadBuffer(MenuPosy,sizeof(Posy));              { read it in         }
  SeekBuffer(MenuPosy.Position);                       { seek the menu      }
end;

{---------------------------------------------------------------------------}

procedure InitialiseMenus;                  { set up the basic menu system  }

Var
  Loop : word;

begin
  for Loop := 1 to MAXSTACK do              { loop over the menu stack      }
    begin
      Menus[Loop] := NIL;                   { ans set the pointer to NIL    }
    end;
  CurrentMenu := 1;                         { this is the root menu         }
  BlockReadBuffer(MenuPosy,sizeof(Posy));   { read in the first menu pos.   }
  SeekBuffer(MenuPosy.Position);            { move to that position         }
  ReadMenu(Menus[CurrentMenu]);             { read in the first menu        }
  MakeWindow(OldWin,                        { open the root window          }
             1,1,80,25,                     { the whole screen              }
             GetCType(MenuHeader.BackColor),{ so the screen is wiped with   }
             GetCType(MenuHeader.BackColor),{ the background color, and it  }
             GetCType(MenuHeader.BackColor),{ has no border                 }
             NOBORDER,'');                  { the old screen is saved       }
end;

{---------------------------------------------------------------------------}

function STI_GetCommand(var Ch : KeyRec; var Cmd : byte) : Boolean;

Var
  Dummy : char;

begin
  STI_GetCommand := TRUE;                   { Assume this is a command      }
  Cmd            := NULLCMD;                { null the command              }
  Ch             := STI_InKey2;             { get the key stroke(s)         }
  case ord(Ch[1]) of
    0  : begin                              { special key pressed           }
           case ord(Ch[2]) of
             63,                            { help or F1                    }
             98 : begin
                    Cmd := HELPCMD;
                  end;
             59 : begin                     { left arrow                    }
                    Cmd := LEFTCMD;
                  end;
             60 : begin                     { right arrow                   }
                    Cmd := RIGHTCMD;
                  end;
             61 : begin                     { down arrow                    }
                    Cmd := DOWNCMD;
                  end;
             58 : begin                     { up arrow                      }
                    Cmd := UPCMD;
                  end;
             else STI_GetCommand := FALSE;
           end;
         end;
    27 : begin                              { escape key pressed            }
           Cmd := ESCAPECMD;
         end;
    13 : begin                              { return key pressed            }
           Cmd := SELECTCMD;
         end;
    19 : begin                              { control S pressed  <-         }
           Cmd := LEFTCMD;
         end;
    4  : begin                              { control D pressed ->          }
           Cmd := RIGHTCMD;
         end;
    5  : begin                              { control E pressed (up)        }
           Cmd := UPCMD;
         end;
    24 : begin                              { control X pressed (down)      }
           Cmd := DOWNCMD;
         end;
    else STI_GetCommand := FALSE;           { this is NOT a command keyset  }
  end;
end;

{---------------------------------------------------------------------------}

procedure SetSelectionNumbers;

Var
  Width      : byte;
  Loop       : word;
  CWidth     : byte;

begin
  CWidth := 0;
  if Menus[CurrentMenu]^.Menu.MenuType = 1 then  { BAR menu                }
    begin
      Width  := (Menus[CurrentMenu]^.Menu.Size[3] -
                 Menus[CurrentMenu]^.Menu.Size[1])-2;
      Loop := Menus[CurrentMenu]^.SS;
      while (CWidth < Width) and (Loop <= Menus[CurrentMenu]^.Menu.SelectNum) do
        begin
          inc(CWidth,length(Menus[CurrentMenu]^.Sels^[Loop].Prompt)+2);
          inc(Loop);
        end;
      if Loop < Menus[CurrentMenu]^.Menu.SelectNum then
        Menus[CurrentMenu]^.SE := Loop-2
      else
        if CWidth > Width then
          Menus[CurrentMenu]^.SE := Loop-2
        else
          if CWidth = Width then
            Menus[CurrentMenu]^.SE := Loop-1
          else
            Menus[CurrentMenu]^.SE  := Menus[CurrentMenu]^.Menu.SelectNum;
    end
  else
    begin
      Width  := (Menus[CurrentMenu]^.Menu.Size[4] -
                 Menus[CurrentMenu]^.Menu.Size[2])-2;
      Loop := Menus[CurrentMenu]^.SS;
      while (CWidth < Width) and (Loop < Menus[CurrentMenu]^.Menu.SelectNum) do
        begin
          inc(CWidth);
          inc(Loop);
        end;
      Menus[CurrentMenu]^.SE  := Loop;
    end;
end;

{---------------------------------------------------------------------------}

procedure DrawItem(X,Y : byte; Prompt : string; HIGH : Boolean);

begin
  if HIGH then
    begin
      GotoXY(X,Y);
      ResetColors;
      TextColor(GetCType(Menus[CurrentMenu]^.Menu.PromptCol));
      Write(Prompt);
    end
  else
    begin
      GotoXY(X,Y);
      ResetColors;
      TextColor(GetCType(Menus[CurrentMenu]^.Menu.TextCol));
      Write(Prompt);
      GotoXY(X,Y);
      ResetColors;
      TextColor(GetCType(Menus[CurrentMenu]^.Menu.HighLight));
      Write(Prompt[1]);
    end;
  ResetColors;
end;

{---------------------------------------------------------------------------}

procedure DrawSelections;

Var
  Loop  : word;
  PosX,
  PosY  : byte;
  PX,PY : byte;

begin
  PosX := Menus[CurrentMenu]^.Menu.Size[1]+1;
  PosY := Menus[CurrentMenu]^.Menu.Size[2]+1;
  PX   := Menus[CurrentMenu]^.Menu.Size[3]-1;
  PY   := Menus[CurrentMenu]^.Menu.Size[4]-1;
  Menus[CurrentMenu]^.X := PosX;
  Menus[CurrentMenu]^.Y := PosY;
  ResetColors;
  window(PosX,PosY,PX,PY);
  TextColor(Menus[CurrentMenu]^.Menu.BodyCol);
  ClrScr;
  ResetColors;
  window(1,1,80,25);
  if Menus[CurrentMenu]^.Menu.MenuType = 1 then  { BAR menu                 }
    begin
      for Loop := Menus[CurrentMenu]^.SS to Menus[CurrentMenu]^.SE do
        begin
          DrawItem(PosX,PosY,Menus[CurrentMenu]^.SelS^[Loop].Prompt,FALSE);
          inc(PosX,length(Menus[CurrentMenu]^.SelS^[Loop].Prompt)+2);
        end;
    end
  else
    begin
      for Loop := Menus[CurrentMenu]^.SS to Menus[CurrentMenu]^.SE do
        begin
          DrawItem(PosX,PosY,Menus[CurrentMenu]^.SelS^[Loop].Prompt,FALSE);
          inc(PosY);
        end;
    end;
end;

{---------------------------------------------------------------------------}

procedure STI_DrawMenu;

begin
  MakeWindow(Menus[CurrentMenu]^.Win,       { open the window               }
             Menus[CurrentMenu]^.Menu.Size[1],
             Menus[CurrentMenu]^.Menu.Size[2],
             Menus[CurrentMenu]^.Menu.Size[3],
             Menus[CurrentMenu]^.Menu.Size[4],
             GetCType(Menus[CurrentMenu]^.Menu.TextCol),
             GetCType(Menus[CurrentMenu]^.Menu.BorderCol),
             GetCType(Menus[CurrentMenu]^.Menu.TitleCol),
             GetBType(Menus[CurrentMenu]^.Menu.BorderType),
             Menus[CurrentMenu]^.Menu.MenuTitle);
  SetSelectionNumbers;
  DrawSelections;
  DrawItem(Menus[CurrentMenu]^.X,
           Menus[CurrentMenu]^.Y,
           Menus[CurrentMenu]^.SelS^[1].Prompt,
           TRUE);
  Menus[CurrentMenu]^.OnScr := TRUE;
end;

{---------------------------------------------------------------------------}

procedure UnDrawMenu;

begin
  DisposeWindow(Menus[CurrentMenu]^.Win);
  freemem(Menus[CurrentMenu]^.SelS,
         Menus[CurrentMenu]^.Menu.SelectNum * sizeof(Sel));
  freemem(Menus[CurrentMenu]^.SelF,
         Menus[CurrentMenu]^.Menu.SelectNum);
  dispose(Menus[CurrentMenu]);
  Menus[CurrentMenu] := NIL;
  dec(CurrentMenu);
end;

{---------------------------------------------------------------------------}

procedure MenuScrollOne;

begin
  if Menus[Currentmenu]^.SS > 1 then
    begin
      dec(Menus[CurrentMenu]^.SS);
      SetSelectionNumbers;
      DrawSelections;
      Menus[CurrentMenu]^.CurSel := Menus[CurrentMenu]^.SS;
      DrawItem(Menus[CurrentMenu]^.X,
               Menus[CurrentMenu]^.Y,
               Menus[CurrentMenu]^.SelS^[Menus[CurrentMenu]^.CurSel].Prompt,
               TRUE);
    end;
end;

{---------------------------------------------------------------------------}

procedure MenuScrollTwo;

begin
  if Menus[Currentmenu]^.SE < Menus[CurrentMenu]^.Menu.SelectNum then
    begin
      Menus[CurrentMenu]^.SS := Menus[CurrentMenu]^.SE;
      SetSelectionNumbers;
      DrawSelections;
      Menus[CurrentMenu]^.CurSel := Menus[CurrentMenu]^.SS;
      DrawItem(Menus[CurrentMenu]^.X,
               Menus[CurrentMenu]^.Y,
               Menus[CurrentMenu]^.SelS^[Menus[CurrentMenu]^.CurSel].Prompt,
               TRUE);
    end;
end;

{---------------------------------------------------------------------------}

procedure UpOrDown(UP : boolean);

begin
  DrawItem(Menus[CurrentMenu]^.X,
           Menus[CurrentMenu]^.Y,
           Menus[CurrentMenu]^.SelS^[Menus[CurrentMenu]^.CurSel].Prompt,
           FALSE);
  if UP then
    begin
      dec(Menus[CurrentMenu]^.CurSel);
      dec(Menus[CurrentMenu]^.Y);
    end
  else
    begin
      inc(Menus[CurrentMenu]^.CurSel);
      inc(Menus[CurrentMenu]^.Y);
    end;
  DrawItem(Menus[CurrentMenu]^.X,
           Menus[CurrentMenu]^.Y,
           Menus[CurrentMenu]^.SelS^[Menus[CurrentMenu]^.CurSel].Prompt,
           TRUE);
end;

{---------------------------------------------------------------------------}

procedure LeftOrRight(LEFT : boolean);

begin
  DrawItem(Menus[CurrentMenu]^.X,
           Menus[CurrentMenu]^.Y,
           Menus[CurrentMenu]^.SelS^[Menus[CurrentMenu]^.CurSel].Prompt,
           FALSE);
  if LEFT then
    begin
      dec(Menus[CurrentMenu]^.CurSel);
      dec(Menus[CurrentMenu]^.X,
          length(Menus[CurrentMenu]^.SelS^[Menus[CurrentMenu]^.CurSel].Prompt)+2);
    end
  else
    begin
      inc(Menus[CurrentMenu]^.X,
          length(Menus[CurrentMenu]^.SelS^[Menus[CurrentMenu]^.CurSel].Prompt)+2);
      inc(Menus[CurrentMenu]^.CurSel);
    end;
  DrawItem(Menus[CurrentMenu]^.X,
           Menus[CurrentMenu]^.Y,
           Menus[CurrentMenu]^.SelS^[Menus[CurrentMenu]^.CurSel].Prompt,
           TRUE);
end;

{---------------------------------------------------------------------------}

procedure STI_EvaluateCommand(Command : byte; var Choice : word; Var ExitFlag : boolean);

var
  Dummy : Boolean;

begin
  ExitFlag := FALSE;
  Choice   := 0;
  case Command of
    NULLCMD    : begin
                 end;
    UPCMD      : begin
                   if Menus[CurrentMenu]^.Menu.MenuType <> 1 then
                     begin
                       if Menus[CurrentMenu]^.CurSel = Menus[CurrentMenu]^.SS then
                         begin
                           MenuScrollOne;
                         end
                       else
                         begin
                           UpOrDown(TRUE);
                         end;
                     end;
                 end;
    DOWNCMD    : begin
                   if Menus[CurrentMenu]^.Menu.MenuType <> 1 then
                     begin
                       if Menus[CurrentMenu]^.CurSel = Menus[CurrentMenu]^.SE then
                         begin
                           MenuScrollTwo;
                         end
                       else
                         begin
                           UpOrDown(FALSE);
                         end;
                     end;
                 end;
    LEFTCMD    : begin
                   case Menus[CurrentMenu]^.Menu.MenuType of
                     1  : begin             { BAR menu                      }
                            if Menus[CurrentMenu]^.CurSel = Menus[CurrentMenu]^.SS then
                              begin
                                MenuScrollOne;
                              end
                            else
                              begin
                                LeftOrRight(TRUE);
                                if Menus[CurrentMenu]^.SelS^[Menus[CurrentMenu]^.CurSel].Jump and
                                   Menus[CurrentMenu]^.ASel then
                                    STI_EvaluateCommand(SELECTCMD,Choice,ExitFlag);
                              end;
                          end;
                     else begin
                            if (Menus[CurrentMenu-1]^.Menu.MenuType = 1)
                                and (Menus[CurrentMenu-1]^.Asel) then
                              begin
                                UnDrawMenu;
                                STI_EvaluateCommand(LEFTCMD,Choice,ExitFlag);
                              end;
                          end;
                   end;
                 end;
    RIGHTCMD   : begin
                   case Menus[CurrentMenu]^.Menu.MenuType of
                     1  : begin             { BAR menu                      }
                            if Menus[CurrentMenu]^.CurSel = Menus[CurrentMenu]^.SE then
                              begin
                                MenuScrollTwo;
                              end
                            else
                              begin
                                LeftOrRight(FALSE);
                                if Menus[CurrentMenu]^.SelS^[Menus[CurrentMenu]^.CurSel].Jump and
                                   Menus[CurrentMenu]^.ASel then
                                    STI_EvaluateCommand(SELECTCMD,Choice,ExitFlag);
                              end;
                          end;
                     else begin
                            if (Menus[CurrentMenu-1]^.Menu.MenuType = 1)
                              and (Menus[CurrentMenu-1]^.Asel) then
                              begin
                                UnDrawMenu;
                                STI_EvaluateCommand(RIGHTCMD,Choice,ExitFlag);
                              end;
                          end;
                   end;
                 end;
    SELECTCMD  : begin
                   Dummy := TRUE;
                   if @VerFunc <> NIL then
                     Dummy := VerFunc(Menus[CurrentMenu]^.Menu.MenuID,Menus[CurrentMenu]^.CurSel);
                   if Dummy then
                     begin
                       if Menus[CurrentMenu]^.SelS^[Menus[CurrentMenu]^.CurSel].Jump = TRUE then
                         begin
                           if Menus[CurrentMenu]^.Menu.MenuType = 1 then
                             Menus[CurrentMenu]^.ASel := TRUE;
                           SeekMenu(Menus[CurrentMenu]^.SelS^[Menus[CurrentMenu]^.CurSel].Value);
                           inc(CurrentMenu);
                           ReadMenu(Menus[CurrentMenu]);
                           STI_DrawMenu;
                         end
                       else
                         begin
                           Choice := Menus[CurrentMenu]^.SelS^[Menus[CurrentMenu]^.CurSel].Value;
                           if (@ExitFunc <> NIL) and ExitFunc(CurrentMenu,Menus[CurrentMenu]^.CurSel) then
                             ExitFlag := TRUE
                           else
                             ExitFlag := FALSE;
                         end;
                     end;
                 end;
    ESCAPECMD  : begin
                   if CurrentMenu <> 1 then
                     UnDrawMenu
                   else
                     if (@ExitFunc <> NIL) and ExitFunc(CurrentMenu,Menus[CurrentMenu]^.CurSel) then
                       begin
                         UnDrawMenu;
                         ExitFlag := TRUE;
                       end
                   else
                     if (@ExitFunc = NIL) then
                       begin
                         UnDrawMenu;
                         ExitFlag := TRUE;
                       end;
                 end;
    HELPCMD    : begin
                   if @HelpProc <> NIL then
                     HelpProc(CurrentMenu,Menus[CurrentMenu]^.CurSel);
                 end;
  end;
end;

{---------------------------------------------------------------------------}

function STI_GetSelection(Inch : KeyRec) : boolean;

var
  Loop  : word;
  Dumm1 : word;
  Dumm2 : boolean;

begin
  STI_GetSelection := FALSE;
  for Loop := Menus[CurrentMenu]^.SS to Menus[CurrentMenu]^.SE do
    begin
      if UpCase(Menus[CurrentMenu]^.SelS^[Loop].Prompt[1]) = UpCase(Inch[1]) then
        begin
          DrawItem(Menus[CurrentMenu]^.X,
                   Menus[CurrentMenu]^.Y,
                   Menus[CurrentMenu]^.SelS^[Menus[CurrentMenu]^.CurSel].Prompt,
                   FALSE);
          if Menus[CurrentMenu]^.Menu.MenuType = 1 then
            Menus[CurrentMenu]^.ASel := FALSE;
          if Menus[CurrentMenu]^.CurSel < Loop then
            begin
              while Menus[CurrentMenu]^.CurSel < Loop do
                begin
                  if Menus[CurrentMenu]^.Menu.MenuType = 1 then
                    STI_EvaluateCommand(RIGHTCMD,Dumm1,Dumm2)
                  else
                    STI_EvaluateCommand(DOWNCMD,Dumm1,Dumm2)
                end;
            end
          else
            begin
              while Menus[CurrentMenu]^.CurSel > Loop do
                begin
                  if Menus[CurrentMenu]^.Menu.MenuType = 1 then
                    STI_EvaluateCommand(LEFTCMD,Dumm1,Dumm2)
                  else
                    STI_EvaluateCommand(UPCMD,Dumm1,Dumm2)
                end;
            end;
          DrawItem(Menus[CurrentMenu]^.X,
                   Menus[CurrentMenu]^.Y,
                   Menus[CurrentMenu]^.SelS^[Menus[CurrentMenu]^.CurSel].Prompt,
                   TRUE);
          STI_GetSelection := TRUE;
          Exit;
        end;
    end;
end;

{---------------------------------------------------------------------------}

function STI_GetMenuChoice(PassOn : boolean; Var ExitFlag : boolean; Var Inch : KeyRec) : word;

var
  Finished : boolean;                       { exit menu flag                }
  Command  : byte;                          { returned keystroke            }
  Choice   : word;                          { the choice returned           }


begin
  HiddenCursor;
  if not Menus[CurrentMenu]^.OnScr then     { is the current menu onscreen? }
    STI_DrawMenu;                           { no, so draw it                }
  Finished := FALSE;                        { not finished yet              }
  repeat                                    { and we loop for a while       }
    begin
      Choice := 0;
      if STI_GetCommand(Inch,Command) then
        begin
          STI_EvaluateCommand(Command,Choice,ExitFlag);
          STI_GetMenuChoice := Choice;
          Finished := (ExitFlag = TRUE) or (Choice > 0);
          if ExitFlag then
            DisposeWindow(OldWin);
        end
      else
        if STI_GetSelection(Inch) then
          begin
            STI_EvaluateCommand(SELECTCMD,Choice,ExitFlag);
            STI_GetMenuChoice := Choice;
            Finished := (ExitFlag = TRUE) or (Choice > 0);
            if ExitFlag then
              DisposeWindow(OldWin);
          end
        else
          begin
            if PassOn then
              begin
                ExitFlag := TRUE;
                Finished := TRUE;
                STI_getMenuChoice := Choice;
              end;
          end;
    end;
  until Finished;                           { until we are finished         }
  NormalCursor;
end;

{---------------------------------------------------------------------------}

function STI_MenusInit(MenuFile : string; HelpRoutine,ValidationRoutine,ExitRoutine,LinkedMenuData : pointer) : shortint;

Label
  JumpTo;

Var
  HP      : HelpProcType  absolute HelpRoutine;
  VP      : VerFuncType   absolute ValidationRoutine;
  EP      : ExitFuncType  absolute ExitRoutine;

begin
  HelpProc := HP;
  VerFunc  := VP;
  ExitFunc := EP;
  MenuDefinition  := NIL;                   { make sure the pointer is nil  }
  InMemory        := TRUE;                  { assume the buffer is in mem.  }
  STI_MenusInit   := STI_NOERROR;           { assume all is ok.             }
  if MenuFile <> '' then                    { menu specification is in file }
    begin
      assign(InFile,MenuFile);              { assign the file name          }
      reset(InFile,1);                      { open the file                 }
      if IOResult <> 0 then                 { is there an error ?           }
        begin
          STI_MenusInit := STI_NOMENUFILE;  { couldn't find the file        }
          close(InFile);                    { set error, close the file     }
          exit;                             { good bye                      }
        end;
      MenuDefSize := filesize(InFile);      { get the file size             }
      if ((MenuDefSize > MAXDEFSIZE) or     { check the file size against   }
         (maxavail < MenuDefSize))   then   { buffer and memory             }
        begin                               { too big, so set a flag and    }
          InMemory := FALSE;                { jump to the block read part   }
          goto JumpTo;
        end;
      getmem(MenuDefinition,MenuDefSize);   { get some memory               }
      blockread(InFile,                     { read in the menu definition   }
                MenuDefinition^[0],
                MenuDefSize);
      if IOResult <> 0 then                 { check for an error            }
        begin
          STI_MenusInit := STI_FILEREADERROR;  { didn't read properly, so   }
          close(InFile);                       { set an error, close file   }
          exit;                             { and return                    }
        end;
      close(InFile);                        { if it wasn't already done     }
    end else
  if LinkedMenuData <> NIL then             { if the data is linked in      }
    begin
      MenuDefinition := LinkedMenuData;     { then we set the pointer to it }
      InMemory := TRUE;
      MenuDefSize := 1024;
    end
  else                                      { no data to play with          }
    begin
      STI_MenusInit := STI_NOMENUFILE;      { set error status and return   }
      exit;                                 { bye bye                       }
    end;
  BufferPos := 0;                           { set position in buffer        }
JumpTo:
  BlockReadBuffer(MenuHeader,sizeof(Head)); { read in the header            }
  if (LinkedMenuData <> NIL) and InMemory then
    MenuDefSize := ord(MenuHeader.Name[length(MenuHeader.Name)])*128;
  if MenuHeader.Mode <> 1 then              { check the mode                }
    begin
      STI_MENUSINIT := STI_MODECLASH;       { there is a mode clash         }
      Exit;
    end;
  InitialiseMenus;                          { set up the menu system proper }
end;

{---------------------------------------------------------------------------}

begin
end.